@@ -32,7 +32,10 @@ from PyQt5 import (
     QtCore,
     QtWidgets,
 )
-from PyQt5.QtNetwork import QNetworkReply
+from PyQt5.QtNetwork import (
+    QNetworkReply,
+    QNetworkRequest,
+)
 
 from picard import (
     config,
@@ -56,9 +59,11 @@ from picard.coverart.utils import (
 )
 from picard.util import webbrowser2
 
+from picard.ui import PicardDialog
 from picard.ui.ui_provider_options_caa import Ui_CaaOptions
 from picard.ui.util import StandardButton
 
+
 CaaSizeItem = namedtuple('CaaSizeItem', ['thumbnail', 'label'])
 
 _CAA_THUMBNAIL_SIZE_MAP = OrderedDict([
@@ -198,7 +203,7 @@ class ListBox(QtWidgets.QListWidget):
             yield self.item(index).data(role)
 
 
-class CAATypesSelectorDialog(QtWidgets.QDialog):
+class CAATypesSelectorDialog(PicardDialog):
     """Display dialog box to select the CAA image types to include and exclude from download and use.
 
     Keyword Arguments:
@@ -215,7 +220,9 @@ class CAATypesSelectorDialog(QtWidgets.QDialog):
             types_exclude = []
 
         self.setWindowTitle(_("Cover art types"))
+        self.setWindowModality(QtCore.Qt.WindowModal)
         self.layout = QtWidgets.QVBoxLayout(self)
+        self.layout.setSizeConstraint(QtWidgets.QLayout.SetFixedSize)
 
         # Create list boxes for dialog
         self.list_include = ListBox()
@@ -544,8 +551,7 @@ class CoverArtProviderCaa(CoverArtProvider):
 
     def enabled(self):
         """Check if CAA artwork has to be downloaded"""
-        if not super().enabled() or \
-                self.coverart.front_image_found:
+        if not super().enabled():
             return False
         if self.restrict_types and not self.len_caa_types:
             log.debug('User disabled all Cover Art Archive types')
@@ -563,7 +569,8 @@ class CoverArtProviderCaa(CoverArtProvider):
             self._caa_path,
             self._caa_json_downloaded,
             priority=True,
-            important=False
+            important=False,
+            cacheloadcontrol=QNetworkRequest.PreferNetwork
         )
         self.album._requests += 1
         # we will call next_in_queue() after json parsing
