@@ -250,8 +250,11 @@ class APEv2File(File):
 
     @classmethod
     def supports_tag(cls, name):
-        return (bool(name) and is_valid_key(name)
-                and name not in UNSUPPORTED_TAGS)
+        return (bool(name) and name not in UNSUPPORTED_TAGS
+                and (is_valid_key(name)
+                    or name.startswith('comment:')
+                    or name.startswith('lyrics:')
+                    or name.startswith('performer:')))
 
 
 class MusepackFile(APEv2File):
