@@ -17,72 +17,68 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 
+from collections import defaultdict
+import re
+from urllib.parse import urlparse
+
+from mutagen import id3
+import mutagen.aiff
 import mutagen.apev2
+import mutagen.dsf
 import mutagen.mp3
 import mutagen.trueaudio
-import re
-from collections import defaultdict
-from mutagen import id3
-from picard import config, log
-from picard.coverart.image import TagCoverArtImage, CoverArtImageError
-from picard.metadata import Metadata
+
+from picard import (
+    config,
+    log,
+)
+from picard.coverart.image import (
+    CoverArtImageError,
+    TagCoverArtImage,
+)
 from picard.file import File
-from picard.formats.mutagenext import compatid3
-from picard.util import encode_filename, sanitize_date
-from urlparse import urlparse
-
-
-# Ugly, but... I need to save the text in ISO-8859-1 even if it contains
-# unsupported characters and this better than encoding, decoding and
-# again encoding.
-def patched_EncodedTextSpec_write(self, config, frame, value):
-    try:
-        enc, term = self._encodings[frame.encoding]
-    except AttributeError:
-        enc, term = self.encodings[frame.encoding]
-    return value.encode(enc, 'ignore') + term
-
-id3._specs.EncodedTextSpec.write = patched_EncodedTextSpec_write
-
-
-# One more "monkey patch". The ID3 spec says that multiple text
-# values should be _separated_ by the string terminator, which
-# means that e.g. 'a\x00' are two values, 'a' and ''.
-def patched_MultiSpec_write(self, config, frame, value):
-    data = self._write_orig(config, frame, value)
-    spec = self.specs[-1]
-    if isinstance(spec, id3._specs.EncodedTextSpec):
-        try:
-            term = spec._encodings[frame.encoding][1]
-        except AttributeError:
-            term = spec.encodings[frame.encoding][1]
-        if data.endswith(term):
-            data = data[:-len(term)]
-    return data
-
-
-id3._specs.MultiSpec._write_orig = id3._specs.MultiSpec.write
-id3._specs.MultiSpec.write = patched_MultiSpec_write
+from picard.formats.mutagenext import (
+    compatid3,
+    delall_ci,
+)
+from picard.metadata import Metadata
+from picard.util import (
+    encode_filename,
+    sanitize_date,
+)
 
 
+id3.GRP1 = compatid3.GRP1
 id3.TCMP = compatid3.TCMP
 id3.TSO2 = compatid3.TSO2
 id3.TSOC = compatid3.TSOC
 
-__ID3_IMAGE_TYPE_MAP = {
-    "other": 0,
-    "obi": 0,
-    "tray": 0,
-    "spine": 0,
-    "sticker": 0,
-    "front": 3,
-    "back": 4,
-    "booklet": 5,
-    "medium": 6,
-    "track": 6,
-}
+__IMAGE_TYPES = [
+    ("obi", 0),
+    ("tray", 0),
+    ("spine", 0),
+    ("sticker", 0),
+    ("other", 0),
+    ("front", 3),
+    ("back", 4),
+    ("booklet", 5),
+    ("track", 6),
+    ("medium", 6),
+]
 
-__ID3_REVERSE_IMAGE_TYPE_MAP = dict([(v, k) for k, v in __ID3_IMAGE_TYPE_MAP.iteritems()])
+__ID3_IMAGE_TYPE_MAP = dict(__IMAGE_TYPES)
+
+__ID3_REVERSE_IMAGE_TYPE_MAP = dict([(v, k) for k, v in __IMAGE_TYPES])
+
+
+def id3text(text, encoding):
+    """Returns a string which only contains code points which can
+    be encododed with the given numeric id3 encoding.
+    """
+
+    if encoding == 0:
+        return text.encode("latin1", "replace").decode("latin1")
+    return text
 
 
 def image_type_from_id3_num(id3type):
@@ -94,24 +90,24 @@ def image_type_as_id3_num(texttype):
 
 
 def types_from_id3(id3type):
-    return [unicode(image_type_from_id3_num(id3type))]
+    return [image_type_from_id3_num(id3type)]
 
 
 class ID3File(File):
 
     """Generic ID3-based file."""
-    _File = None
     _IsMP3 = False
 
     __upgrade = {
         'XSOP': 'TSOP',
         'TXXX:ALBUMARTISTSORT': 'TSO2',
         'TXXX:COMPOSERSORT': 'TSOC',
+        'TXXX:mood': 'TMOO',
     }
 
     __translate = {
         # In same sequence as defined at http://id3.org/id3v2.4.0-frames
-        'TIT1': 'grouping',
+        # 'TIT1': 'grouping', # Depends on itunes_compatible_grouping
         'TIT2': 'title',
         'TIT3': 'subtitle',
         'TALB': 'album',
@@ -125,6 +121,7 @@ class ID3File(File):
         'TCOM': 'composer',
         'TENC': 'encodedby',
         'TBPM': 'bpm',
+        'TKEY': 'key',
         'TLAN': 'language',
         'TCON': 'genre',
         'TMED': 'media',
@@ -140,13 +137,17 @@ class ID3File(File):
         'WCOP': 'license',
         'WOAR': 'website',
         'COMM': 'comment',
+        'TOAL': 'originalalbum',
+        'TOPE': 'originalartist',
 
         # The following are informal iTunes extensions to id3v2:
         'TCMP': 'compilation',
         'TSOC': 'composersort',
         'TSO2': 'albumartistsort',
+        'MVNM': 'movement'
     }
-    __rtranslate = dict([(v, k) for k, v in __translate.iteritems()])
+    __rtranslate = dict([(v, k) for k, v in __translate.items()])
+    __translate['GRP1'] = 'grouping'  # Always read, but writing depends on itunes_compatible_grouping
 
     __translate_freetext = {
         'MusicBrainz Artist Id': 'musicbrainz_artistid',
@@ -159,6 +160,8 @@ class ID3File(File):
         'MusicBrainz Disc Id': 'musicbrainz_discid',
         'MusicBrainz Work Id': 'musicbrainz_workid',
         'MusicBrainz Release Group Id': 'musicbrainz_releasegroupid',
+        'MusicBrainz Original Album Id': 'musicbrainz_originalalbumid',
+        'MusicBrainz Original Artist Id': 'musicbrainz_originalartistid',
         'MusicBrainz Album Release Country': 'releasecountry',
         'MusicIP PUID': 'musicip_puid',
         'Acoustid Fingerprint': 'acoustid_fingerprint',
@@ -169,13 +172,33 @@ class ID3File(File):
         'BARCODE': 'barcode',
         'ASIN': 'asin',
         'MusicMagic Fingerprint': 'musicip_fingerprint',
-        'Artists': 'artists',
-        'Work': 'work',
+        'ARTISTS': 'artists',
+        'WORK': 'work',
         'Writer': 'writer',
+        'SHOWMOVEMENT': 'showmovement',
     }
-    __rtranslate_freetext = dict([(v, k) for k, v in __translate_freetext.iteritems()])
+    __rtranslate_freetext = dict([(v, k) for k, v in __translate_freetext.items()])
     __translate_freetext['writer'] = 'writer'  # For backward compatibility of case
 
+    # Freetext fields that are loaded case-insensitive
+    __rtranslate_freetext_ci = {
+        'replaygain_album_gain': 'REPLAYGAIN_ALBUM_GAIN',
+        'replaygain_album_peak': 'REPLAYGAIN_ALBUM_PEAK',
+        'replaygain_album_range': 'REPLAYGAIN_ALBUM_RANGE',
+        'replaygain_track_gain': 'REPLAYGAIN_TRACK_GAIN',
+        'replaygain_track_peak': 'REPLAYGAIN_TRACK_PEAK',
+        'replaygain_track_range': 'REPLAYGAIN_TRACK_RANGE',
+        'replaygain_reference_loudness': 'REPLAYGAIN_REFERENCE_LOUDNESS',
+    }
+    __translate_freetext_ci = dict([(b.lower(), a) for a, b in __rtranslate_freetext_ci.items()])
+
+    # Obsolete tag names which will still be loaded, but will get renamed on saving
+    __rename_freetext = {
+        'Artists': 'ARTISTS',
+        'Work': 'WORK',
+    }
+    __rrename_freetext = dict([(v, k) for k, v in __rename_freetext.items()])
+
     _tipl_roles = {
         'engineer': 'engineer',
         'arranger': 'arranger',
@@ -183,18 +206,35 @@ class ID3File(File):
         'DJ-mix': 'djmixer',
         'mix': 'mixer',
     }
-    _rtipl_roles = dict([(v, k) for k, v in _tipl_roles.iteritems()])
+    _rtipl_roles = dict([(v, k) for k, v in _tipl_roles.items()])
 
     __other_supported_tags = ("discnumber", "tracknumber",
-                              "totaldiscs", "totaltracks")
+                              "totaldiscs", "totaltracks",
+                              "movementnumber", "movementtotal")
     __tag_re_parse = {
         'TRCK': re.compile(r'^(?P<tracknumber>\d+)(?:/(?P<totaltracks>\d+))?$'),
-        'TPOS': re.compile(r'^(?P<discnumber>\d+)(?:/(?P<totaldiscs>\d+))?$')
+        'TPOS': re.compile(r'^(?P<discnumber>\d+)(?:/(?P<totaldiscs>\d+))?$'),
+        'MVIN': re.compile(r'^(?P<movementnumber>\d+)(?:/(?P<movementtotal>\d+))?$')
     }
 
+    def __init__(self, filename):
+        super().__init__(filename)
+        self.__casemap = {}
+
+    def build_TXXX(self, encoding, desc, values):
+        """Construct and return a TXXX frame."""
+        # This is here so that plugins can customize the behavior of TXXX
+        # frames in particular via subclassing.
+        # discussion: https://github.com/metabrainz/picard/pull/634
+        # discussion: https://github.com/metabrainz/picard/pull/635
+        # Used in the plugin "Compatible TXXX frames"
+        # PR: https://github.com/metabrainz/picard-plugins/pull/83
+        return id3.TXXX(encoding=encoding, desc=desc, text=values)
+
     def _load(self, filename):
         log.debug("Loading file %r", filename)
-        file = self._File(encode_filename(filename), ID3=compatid3.CompatID3)
+        self.__casemap = {}
+        file = self._get_file(encode_filename(filename))
         tags = file.tags or {}
         # upgrade custom 2.3 frames to 2.4
         for old, new in self.__upgrade.items():
@@ -206,16 +246,22 @@ class ID3File(File):
             frameid = frame.FrameID
             if frameid in self.__translate:
                 name = self.__translate[frameid]
-                if frameid.startswith('T'):
+                if frameid.startswith('T') or frameid in ["GRP1", "MVNM"]:
                     for text in frame.text:
                         if text:
-                            metadata.add(name, unicode(text))
+                            metadata.add(name, str(text))
                 elif frameid == 'COMM':
                     for text in frame.text:
                         if text:
-                            metadata.add('%s:%s' % (name, frame.desc), unicode(text))
+                            metadata.add('%s:%s' % (name, frame.desc), str(text))
                 else:
-                    metadata.add(name, unicode(frame))
+                    metadata.add(name, str(frame))
+            elif frameid == 'TIT1':
+                itunes_compatible = config.setting['itunes_compatible_grouping']
+                name = 'work' if itunes_compatible else 'grouping'
+                for text in frame.text:
+                    if text:
+                        metadata.add(name, str(text))
             elif frameid == "TMCL":
                 for role, name in frame.people:
                     if role or name:
@@ -230,10 +276,17 @@ class ID3File(File):
                         metadata.add('performer:%s' % role, name)
             elif frameid == 'TXXX':
                 name = frame.desc
-                if name in self.__translate_freetext:
+                name_lower = name.lower()
+                if name in self.__rename_freetext:
+                    name = self.__rename_freetext[name]
+                if name_lower in self.__translate_freetext_ci:
+                    orig_name = name
+                    name = self.__translate_freetext_ci[name_lower]
+                    self.__casemap[name] = orig_name
+                elif name in self.__translate_freetext:
                     name = self.__translate_freetext[name]
-                elif ((name in self.__rtranslate) !=
-                        (name in self.__rtranslate_freetext)):
+                elif ((name in self.__rtranslate)
+                      != (name in self.__rtranslate_freetext)):
                     # If the desc of a TXXX frame conflicts with the name of a
                     # Picard tag, load it into ~id3:TXXX:desc rather than desc.
                     #
@@ -243,18 +296,18 @@ class ID3File(File):
                     # ways.) Currently, the only tag in both is license.
                     name = '~id3:TXXX:' + name
                 for text in frame.text:
-                    metadata.add(name, unicode(text))
+                    metadata.add(name, str(text))
             elif frameid == 'USLT':
                 name = 'lyrics'
                 if frame.desc:
                     name += ':%s' % frame.desc
-                metadata.add(name, unicode(frame.text))
+                metadata.add(name, str(frame.text))
             elif frameid == 'UFID' and frame.owner == 'http://musicbrainz.org':
                 metadata['musicbrainz_recordingid'] = frame.data.decode('ascii', 'ignore')
             elif frameid in self.__tag_re_parse.keys():
                 m = self.__tag_re_parse[frameid].search(frame.text[0])
                 if m:
-                    for name, value in m.groupdict().iteritems():
+                    for name, value in m.groupdict().items():
                         if value is not None:
                             metadata[name] = value
                 else:
@@ -272,11 +325,11 @@ class ID3File(File):
                 except CoverArtImageError as e:
                     log.error('Cannot load image from %r: %s' % (filename, e))
                 else:
-                    metadata.append_image(coverartimage)
+                    metadata.images.append(coverartimage)
             elif frameid == 'POPM':
                 # Rating in ID3 ranges from 0 to 255, normalize this to the range 0 to 5
                 if frame.email == config.setting['rating_user_email']:
-                    rating = unicode(int(round(frame.rating / 255.0 * (config.setting['rating_steps'] - 1))))
+                    rating = str(int(round(frame.rating / 255.0 * (config.setting['rating_steps'] - 1))))
                     metadata.add('~rating', rating)
 
         if 'date' in metadata:
@@ -290,20 +343,14 @@ class ID3File(File):
     def _save(self, filename, metadata):
         """Save metadata to the file."""
         log.debug("Saving file %r", filename)
-        try:
-            tags = compatid3.CompatID3(encode_filename(filename))
-        except mutagen.id3.ID3NoHeaderError:
-            tags = compatid3.CompatID3()
+        tags = self._get_tags(filename)
 
         if config.setting['clear_existing_tags']:
             tags.clear()
-        if metadata.images_to_be_saved_to_tags:
+        images_to_save = list(metadata.images.to_be_saved_to_tags())
+        if images_to_save:
             tags.delall('APIC')
 
-        if config.setting['write_id3v1']:
-            v1 = 2
-        else:
-            v1 = 0
         encoding = {'utf-8': 3, 'utf-16': 1}.get(config.setting['id3v2_encoding'], 0)
 
         if 'tracknumber' in metadata:
@@ -311,21 +358,28 @@ class ID3File(File):
                 text = '%s/%s' % (metadata['tracknumber'], metadata['totaltracks'])
             else:
                 text = metadata['tracknumber']
-            tags.add(id3.TRCK(encoding=0, text=text))
+            tags.add(id3.TRCK(encoding=0, text=id3text(text, 0)))
 
         if 'discnumber' in metadata:
             if 'totaldiscs' in metadata:
                 text = '%s/%s' % (metadata['discnumber'], metadata['totaldiscs'])
             else:
                 text = metadata['discnumber']
-            tags.add(id3.TPOS(encoding=0, text=text))
+            tags.add(id3.TPOS(encoding=0, text=id3text(text, 0)))
+
+        if 'movementnumber' in metadata:
+            if 'movementtotal' in metadata:
+                text = '%s/%s' % (metadata['movementnumber'], metadata['movementtotal'])
+            else:
+                text = metadata['movementnumber']
+            tags.add(id3.MVIN(encoding=0, text=id3text(text, 0)))
 
         # This is necessary because mutagens HashKey for APIC frames only
         # includes the FrameID (APIC) and description - it's basically
         # impossible to save two images, even of different types, without
         # any description.
         counters = defaultdict(lambda: 0)
-        for image in metadata.images_to_be_saved_to_tags:
+        for image in images_to_save:
             desc = desctag = image.comment
             if counters[desc] > 0:
                 if desc:
@@ -336,23 +390,30 @@ class ID3File(File):
             tags.add(id3.APIC(encoding=0,
                               mime=image.mimetype,
                               type=image_type_as_id3_num(image.maintype),
-                              desc=desctag,
+                              desc=id3text(desctag, 0),
                               data=image.data))
 
         tmcl = mutagen.id3.TMCL(encoding=encoding, people=[])
         tipl = mutagen.id3.TIPL(encoding=encoding, people=[])
-
-        tags.delall('TCMP')
         for name, values in metadata.rawitems():
-            if name.startswith('performer:'):
+            values = [id3text(v, encoding) for v in values]
+            name = id3text(name, encoding)
+
+            if not self.supports_tag(name):
+                continue
+            elif name.startswith('performer:'):
                 role = name.split(':', 1)[1]
                 for value in values:
-                    tmcl.people.append([role, value])
+                    if config.setting['write_id3v23']:
+                        # TIPL will be upgraded to IPLS
+                        tipl.people.append([role, value])
+                    else:
+                        tmcl.people.append([role, value])
             elif name.startswith('comment:'):
                 desc = name.split(':', 1)[1]
-                if desc.lower()[:4] == "itun":
+                if desc.lower()[:4] == 'itun':
                     tags.delall('COMM:' + desc)
-                    tags.add(id3.COMM(encoding=0, desc=desc, lang='eng', text=[v + u'\x00' for v in values]))
+                    tags.add(id3.COMM(encoding=0, desc=desc, lang='eng', text=[v + '\x00' for v in values]))
                 else:
                     tags.add(id3.COMM(encoding=encoding, desc=desc, lang='eng', text=values))
             elif name.startswith('lyrics:') or name == 'lyrics':
@@ -366,7 +427,7 @@ class ID3File(File):
                 for value in values:
                     tipl.people.append([self._rtipl_roles[name], value])
             elif name == 'musicbrainz_recordingid':
-                tags.add(id3.UFID(owner='http://musicbrainz.org', data=str(values[0])))
+                tags.add(id3.UFID(owner='http://musicbrainz.org', data=bytes(values[0], 'ascii')))
             elif name == '~rating':
                 # Search for an existing POPM frame to get the current playcount
                 for frame in tags.values():
@@ -379,6 +440,15 @@ class ID3File(File):
                 # Convert rating to range between 0 and 255
                 rating = int(round(float(values[0]) * 255 / (config.setting['rating_steps'] - 1)))
                 tags.add(id3.POPM(email=config.setting['rating_user_email'], rating=rating, count=count))
+            elif name == 'grouping':
+                if config.setting['itunes_compatible_grouping']:
+                    tags.add(id3.GRP1(encoding=encoding, text=values))
+                else:
+                    tags.add(id3.TIT1(encoding=encoding, text=values))
+            elif name == 'work' and config.setting['itunes_compatible_grouping']:
+                tags.add(id3.TIT1(encoding=encoding, text=values))
+                tags.delall('TXXX:Work')
+                tags.delall('TXXX:WORK')
             elif name in self.__rtranslate:
                 frameid = self.__rtranslate[name]
                 if frameid.startswith('W'):
@@ -386,13 +456,18 @@ class ID3File(File):
                     if frameid == 'WCOP':
                         # Only add WCOP if there is only one license URL, otherwise use TXXX:LICENSE
                         if len(values) > 1 or not valid_urls:
-                            tags.add(id3.TXXX(encoding=encoding, desc=self.__rtranslate_freetext[name], text=values))
+                            tags.add(self.build_TXXX(encoding, self.__rtranslate_freetext[name], values))
                         else:
                             tags.add(id3.WCOP(url=values[0]))
                     elif frameid == 'WOAR' and valid_urls:
                         for url in values:
                             tags.add(id3.WOAR(url=url))
-                elif frameid.startswith('T'):
+                elif frameid.startswith('T') or frameid == 'MVNM':
+                    if config.setting['write_id3v23']:
+                        if frameid == 'TMOO':
+                            tags.add(self.build_TXXX(encoding, 'mood', values))
+                    # No need to care about the TMOO tag being added again as it is
+                    # automatically deleted by Mutagen if id2v23 is selected
                     tags.add(getattr(id3, frameid)(encoding=encoding, text=values))
                     if frameid == 'TSOA':
                         tags.delall('XSOA')
@@ -400,43 +475,147 @@ class ID3File(File):
                         tags.delall('XSOP')
                     elif frameid == 'TSO2':
                         tags.delall('TXXX:ALBUMARTISTSORT')
+            elif name in self.__rtranslate_freetext_ci:
+                if name in self.__casemap:
+                    description = self.__casemap[name]
+                else:
+                    description = self.__rtranslate_freetext_ci[name]
+                delall_ci(tags, 'TXXX:' + description)
+                tags.add(self.build_TXXX(encoding, description, values))
             elif name in self.__rtranslate_freetext:
-                tags.add(id3.TXXX(encoding=encoding, desc=self.__rtranslate_freetext[name], text=values))
+                description = self.__rtranslate_freetext[name]
+                if description in self.__rrename_freetext:
+                    tags.delall('TXXX:' + self.__rrename_freetext[description])
+                tags.add(self.build_TXXX(encoding, description, values))
             elif name.startswith('~id3:'):
                 name = name[5:]
                 if name.startswith('TXXX:'):
-                    tags.add(id3.TXXX(encoding=encoding, desc=name[5:], text=values))
+                    tags.add(self.build_TXXX(encoding, name[5:], values))
                 else:
                     frameclass = getattr(id3, name[:4], None)
                     if frameclass:
                         tags.add(frameclass(encoding=encoding, text=values))
             # don't save private / already stored tags
-            elif not name.startswith("~") and not name in self.__other_supported_tags:
-                tags.add(id3.TXXX(encoding=encoding, desc=name, text=values))
+            elif not name.startswith("~") and name not in self.__other_supported_tags:
+                tags.add(self.build_TXXX(encoding, name, values))
 
         if tmcl.people:
             tags.add(tmcl)
         if tipl.people:
             tags.add(tipl)
 
-        if config.setting['write_id3v23']:
-            tags.update_to_v23(join_with=config.setting['id3v23_join_with'])
-            tags.save(encode_filename(filename), v2=3, v1=v1)
-        else:
-            tags.update_to_v24()
-            tags.save(encode_filename(filename), v2=4, v1=v1)
+        self._remove_deleted_tags(metadata, tags)
+
+        self._save_tags(tags, encode_filename(filename))
 
         if self._IsMP3 and config.setting["remove_ape_from_mp3"]:
             try:
                 mutagen.apev2.delete(encode_filename(filename))
-            except:
+            except BaseException:
                 pass
 
-    def supports_tag(self, name):
-        return name in self.__rtranslate or name in self.__rtranslate_freetext\
-            or name.startswith('performer:')\
-            or name.startswith('lyrics:') or name == 'lyrics'\
-            or name in self.__other_supported_tags
+    def _remove_deleted_tags(self, metadata, tags):
+        """Remove the tags from the file that were deleted in the UI"""
+        for name in metadata.deleted_tags:
+            real_name = self._get_tag_name(name)
+            try:
+                if name.startswith('performer:'):
+                    role = name.split(':', 1)[1]
+                    for key, frame in tags.items():
+                        if frame.FrameID in ('TMCL', 'TIPL', 'IPLS'):
+                            for people in frame.people:
+                                if people[0] == role:
+                                    frame.people.remove(people)
+                elif name.startswith('comment:'):
+                    desc = name.split(':', 1)[1]
+                    if desc.lower()[:4] != 'itun':
+                        for key, frame in list(tags.items()):
+                            if frame.FrameID == 'COMM' and frame.desc == desc:
+                                del tags[key]
+                elif name.startswith('lyrics:') or name == 'lyrics':
+                    if ':' in name:
+                        desc = name.split(':', 1)[1]
+                    else:
+                        desc = ''
+                    for key, frame in list(tags.items()):
+                        if frame.FrameID == desc:
+                            del tags[key]
+                elif name in self._rtipl_roles:
+                    role = self._rtipl_roles[name]
+                    for key, frame in tags.items():
+                        if frame.FrameID in ('TIPL', 'IPLS'):
+                            for people in frame.people:
+                                if people[0] == role:
+                                    frame.people.remove(people)
+                elif name == 'musicbrainz_recordingid':
+                    for key, frame in list(tags.items()):
+                        if frame.FrameID == 'UFID' and frame.owner == 'http://musicbrainz.org':
+                            del tags[key]
+                elif real_name == 'POPM':
+                    for key, frame in list(tags.items()):
+                        if frame.FrameID == 'POPM' and frame.email == config.setting['rating_user_email']:
+                            del tags[key]
+                elif real_name in self.__translate:
+                    del tags[real_name]
+                elif real_name in self.__translate_freetext:
+                    tags.delall('TXXX:' + real_name)
+                    if real_name in self.__rrename_freetext:
+                        tags.delall('TXXX:' + self.__rrename_freetext[real_name])
+                elif not name.startswith("~") and name not in self.__other_supported_tags:
+                    tags.delall('TXXX:' + name)
+                elif name.startswith("~"):
+                    name = name[1:]
+                    tags.delall('TXXX:' + name)
+                elif name in self.__other_supported_tags:
+                    del tags[real_name]
+            except KeyError:
+                pass
+
+    @classmethod
+    def supports_tag(cls, name):
+        unsupported_tags = ['r128_album_gain', 'r128_track_gain']
+        return ((name and not name.startswith("~") and name not in unsupported_tags)
+                or name in ("~rating", "~length")
+                or name.startswith("~id3"))
+
+    def _get_tag_name(self, name):
+        if name in self.__rtranslate:
+            return self.__rtranslate[name]
+        elif name in self.__rtranslate_freetext:
+            return self.__rtranslate_freetext[name]
+        elif name == '~rating':
+            return 'POPM'
+        elif name == 'tracknumber':
+            return 'TRCK'
+        elif name == 'discnumber':
+            return 'TPOS'
+        elif name == 'movementnumber':
+            return 'MVIN'
+        else:
+            return None
+
+    def _get_file(self, filename):
+        raise NotImplementedError()
+
+    def _get_tags(self, filename):
+        try:
+            return compatid3.CompatID3(encode_filename(filename))
+        except mutagen.id3.ID3NoHeaderError:
+            return compatid3.CompatID3()
+
+    def _save_tags(self, tags, filename):
+        if config.setting['write_id3v1']:
+            v1 = 2
+        else:
+            v1 = 0
+
+        if config.setting['write_id3v23']:
+            tags.update_to_v23()
+            separator = config.setting['id3v23_join_with']
+            tags.save(filename, v2_version=3, v1=v1, v23_sep=separator)
+        else:
+            tags.update_to_v24()
+            tags.save(filename, v2_version=4, v1=v1)
 
     @property
     def new_metadata(self):
@@ -462,7 +641,7 @@ class ID3File(File):
             # If this is a multi-valued field, then it needs to be flattened,
             # unless it's TIPL or TMCL which can still be multi-valued.
 
-            if (len(values) > 1 and not name in ID3File._rtipl_roles
+            if (len(values) > 1 and name not in ID3File._rtipl_roles
                     and not name.startswith("performer:")):
                 values = [join_with.join(values)]
 
@@ -474,13 +653,16 @@ class ID3File(File):
 class MP3File(ID3File):
 
     """MP3 file."""
-    EXTENSIONS = [".mp3", ".mp2"]
+    EXTENSIONS = [".mp3", ".mp2", ".m2a"]
     NAME = "MPEG-1 Audio"
-    _File = mutagen.mp3.MP3
     _IsMP3 = True
+    _File = mutagen.mp3.MP3
+
+    def _get_file(self, filename):
+        return self._File(filename, ID3=compatid3.CompatID3)
 
     def _info(self, metadata, file):
-        super(MP3File, self)._info(metadata, file)
+        super()._info(metadata, file)
         id3version = ''
         if file.tags is not None and file.info.layer == 3:
             id3version = ' - ID3v%d.%d' % (file.tags.version[0], file.tags.version[1])
@@ -494,6 +676,47 @@ class TrueAudioFile(ID3File):
     NAME = "The True Audio"
     _File = mutagen.trueaudio.TrueAudio
 
-    def _info(self, metadata, file):
-        super(TrueAudioFile, self)._info(metadata, file)
-        metadata['~format'] = self.NAME
+    def _get_file(self, filename):
+        return self._File(filename, ID3=compatid3.CompatID3)
+
+
+class DSFFile(ID3File):
+
+    """DSF file."""
+    EXTENSIONS = [".dsf", '.dff']
+    NAME = "DSF"
+    _File = mutagen.dsf.DSF
+
+    def _get_file(self, filename):
+        return self._File(filename, known_frames=compatid3.known_frames)
+
+    def _get_tags(self, filename):
+        file = self._get_file(filename)
+        if file.tags is None:
+            file.add_tags()
+        return file.tags
+
+    def _save_tags(self, tags, filename):
+        if config.setting['write_id3v23']:
+            tags.update_to_v23()
+            separator = config.setting['id3v23_join_with']
+            tags.save(filename, v2_version=3, v23_sep=separator)
+        else:
+            tags.update_to_v24()
+            tags.save(filename, v2_version=4)
+
+    @classmethod
+    def supports_tag(cls, name):
+        return (super().supports_tag(name)
+                and name not in {'albumsort',
+                                 'artistsort',
+                                 'discsubtitle',
+                                 'titlesort'})
+
+
+class AiffFile(DSFFile):
+
+    """AIFF file."""
+    EXTENSIONS = [".aiff", ".aif", ".aifc"]
+    NAME = "Audio Interchange File Format (AIFF)"
+    _File = mutagen.aiff.AIFF
