@@ -21,17 +21,22 @@ import builtins
 import gettext
 import locale
 import os.path
-import sys
 
 from PyQt5.QtCore import QLocale
 
+from picard.const.sys import (
+    IS_MACOS,
+    IS_WIN,
+)
+
+
 builtins.__dict__['N_'] = lambda a: a
 
 
 def setup_gettext(localedir, ui_language=None, logger=None):
     """Setup locales, load translations, install gettext functions."""
     if not logger:
-        logger = lambda *a, **b: None  # noop
+        logger = lambda *a, **b: None  # noqa: E731
     current_locale = ''
     if ui_language:
         try:
@@ -40,7 +45,7 @@ def setup_gettext(localedir, ui_language=None, logger=None):
         except Exception as e:
             logger(e)
     else:
-        if sys.platform == 'win32':
+        if IS_WIN:
             from ctypes import windll
             try:
                 current_locale = locale.windows_locale[windll.kernel32.GetUserDefaultUILanguage()]
@@ -53,17 +58,19 @@ def setup_gettext(localedir, ui_language=None, logger=None):
                     logger(e)
             except Exception as e:
                 logger(e)
-        elif sys.platform == 'darwin':
+        elif IS_MACOS:
             try:
                 import Foundation
                 defaults = Foundation.NSUserDefaults.standardUserDefaults()
                 current_locale = defaults.objectForKey_('AppleLanguages')[0]
+                current_locale = current_locale.replace('-', '_')
                 locale.setlocale(locale.LC_ALL, current_locale)
             except Exception as e:
                 logger(e)
         else:
             try:
-                current_locale = locale.setlocale(locale.LC_ALL, '')
+                locale.setlocale(locale.LC_ALL, '')
+                current_locale = '.'.join(locale.getlocale(locale.LC_MESSAGES))
             except Exception as e:
                 logger(e)
     os.environ['LANGUAGE'] = os.environ['LANG'] = current_locale
@@ -72,7 +79,7 @@ def setup_gettext(localedir, ui_language=None, logger=None):
     try:
         logger("Loading gettext translation, localedir=%r", localedir)
         trans = gettext.translation("picard", localedir)
-        trans.install(True)
+        trans.install()
         _ngettext = trans.ngettext
         logger("Loading gettext translation (picard-countries), localedir=%r", localedir)
         trans_countries = gettext.translation("picard-countries", localedir)
