@@ -114,6 +114,9 @@ class PluginTreeWidgetItem(HashableTreeWidgetItem):
     @staticmethod
     def set_icon(button, stdicon):
         button.setIcon(button.style().standardIcon(getattr(QtWidgets.QStyle, stdicon)))
+        # Workaround for Qt sometimes not updating the icon.
+        # See https://tickets.metabrainz.org/browse/PICARD-1647
+        button.repaint()
 
     def show_install(self, button, mode):
         if mode == 'hide':
@@ -388,7 +391,7 @@ class PluginsOptionsPage(OptionsPage):
         QtWidgets.QMessageBox.critical(
             self,
             _("Plugin '%s'") % plugin_name,
-            _("An error occured while loading the plugin '%s':\n\n%s") % (plugin_name, error)
+            _("An error occurred while loading the plugin '%s':\n\n%s") % (plugin_name, error)
         )
 
     def plugin_installed(self, plugin):
